# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import os
import pickle
import re

from g2p_en import G2p

from parrots.symbols import symbols

pwd_path = os.path.abspath(os.path.dirname(__file__))
CMU_DICT_PATH = os.path.join(pwd_path, 'data', "cmudict.rep")
CACHE_PATH = os.path.join(pwd_path, 'data', "cmudict_cache.pickle")
_g2p = G2p()

arpa = {
    "AH0",
    "S",
    "AH1",
    "EY2",
    "AE2",
    "EH0",
    "OW2",
    "UH0",
    "NG",
    "B",
    "G",
    "AY0",
    "M",
    "AA0",
    "F",
    "AO0",
    "ER2",
    "UH1",
    "IY1",
    "AH2",
    "DH",
    "IY0",
    "EY1",
    "IH0",
    "K",
    "N",
    "W",
    "IY2",
    "T",
    "AA1",
    "ER1",
    "EH2",
    "OY0",
    "UH2",
    "UW1",
    "Z",
    "AW2",
    "AW1",
    "V",
    "UW2",
    "AA2",
    "ER",
    "AW0",
    "UW0",
    "R",
    "OW1",
    "EH1",
    "ZH",
    "AE0",
    "IH2",
    "IH",
    "Y",
    "JH",
    "P",
    "AY1",
    "EY0",
    "OY2",
    "TH",
    "HH",
    "D",
    "ER0",
    "CH",
    "AO1",
    "AE1",
    "AO2",
    "OY1",
    "AY2",
    "IH1",
    "OW0",
    "L",
    "SH",
}


def replace_phs(phs):
    rep_map = {";": ",", ":": ",", "'": "-", '"': "-"}
    phs_new = []
    for ph in phs:
        if ph in symbols:
            phs_new.append(ph)
        elif ph in rep_map.keys():
            phs_new.append(rep_map[ph])
        else:
            print("ph not in symbols: ", ph)
    return phs_new


def read_dict():
    g2p_dict = {}
    start_line = 49
    with open(CMU_DICT_PATH) as f:
        line = f.readline()
        line_index = 1
        while line:
            if line_index >= start_line:
                line = line.strip()
                word_split = line.split("  ")
                word = word_split[0]

                syllable_split = word_split[1].split(" - ")
                g2p_dict[word] = []
                for syllable in syllable_split:
                    phone_split = syllable.split(" ")
                    g2p_dict[word].append(phone_split)

            line_index = line_index + 1
            line = f.readline()

    return g2p_dict


def cache_dict(g2p_dict, file_path):
    with open(file_path, "wb") as pickle_file:
        pickle.dump(g2p_dict, pickle_file)


def get_dict():
    if os.path.exists(CACHE_PATH):
        with open(CACHE_PATH, "rb") as pickle_file:
            g2p_dict = pickle.load(pickle_file)
    else:
        g2p_dict = read_dict()
        cache_dict(g2p_dict, CACHE_PATH)

    return g2p_dict


eng_dict = get_dict()


def text_normalize(text):
    # todo: eng text normalize
    return text.replace(";", ",")


def g2p(text):
    phones = []
    words = re.split(r"([,;.\-\?\!\s+])", text)
    for w in words:
        if w.upper() in eng_dict:
            phns = eng_dict[w.upper()]
            for ph in phns:
                phones += ph
        else:
            phone_list = list(filter(lambda p: p != " ", _g2p(w)))
            for ph in phone_list:
                if ph in arpa:
                    phones.append(ph)
                else:
                    phones.append(ph)

    return replace_phs(phones)


if __name__ == "__main__":
    print(g2p("hello"))
    print(g2p("In this; paper, we propose 1 DSPGAN, a GAN-based universal vocoder."))



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import argparse
import json
import os
import re
import struct
import sys
import wave
from enum import Enum
from typing import Union, Optional

import librosa
import numpy as np
import soundfile
import torch
import nltk
from huggingface_hub import snapshot_download
from transformers import AutoModelForMaskedLM, AutoTokenizer

from parrots.log import logger
try:
    import parrots.utils as _parrots_utils
    sys.modules.setdefault('utils', _parrots_utils)
except Exception as e:
    logger.debug(f"Failed to alias utils -> parrots.utils: {e}")

from parrots import cnhubert
from parrots.mel_processing import spectrogram_torch
from parrots.synthesizer_model import SynthesizerModel
from parrots.t2s_model import Text2SemanticDecoder
from parrots.text_utils import clean_text, cleaned_text_to_sequence
from parrots.symbols import sentence_split_symbols


os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"
os.environ["TOKENIZERS_PARALLELISM"] = "TRUE"

# Constants for speaker model
CONFIG_NAME = "config.json"
SOVITS_MODEL_NAME = "sovits.pth"
GPT_MODEL_NAME = "gpt.ckpt"
REF_WAV_NAME = "ref.wav"
SPEAKER_NAMES = ["MaiMai", "XingTong", "XuanShen", "KusanagiNene", "LongShouRen", "KuileBlanc"]


class LANG(Enum):
    auto = 0  # 多语种按句子切分识别各自语种
    zh = 1  # 全部按中文识别
    en = 2  # 全部按英文识别
    ja = 3  # 全部按日文识别

    def __str__(self):
        return self.name

    @staticmethod
    def from_string(s):
        try:
            return LANG[s]
        except KeyError:
            raise ValueError()


def load_audio(file, sr):
    """
    Load audio file and resample to target sample rate.
    First try using ffmpeg, if failed, fallback to librosa or soundfile.
    
    Args:
        file: path to audio file
        sr: target sample rate
    
    Returns:
        audio: numpy array of audio samples
    """
    file = (
        file.strip(" ").strip('"').strip("\n").strip('"').strip(" ")
    )  # 防止小白拷路径头尾带了空格和"和回车
    
    # Try ffmpeg first (fastest and most robust)
    try:
        import ffmpeg
        # https://github.com/openai/whisper/blob/main/whisper/audio.py#L26
        # This launches a subprocess to decode audio while down-mixing and resampling as necessary.
        # Requires the ffmpeg CLI and `ffmpeg-python` package to be installed.
        out, _ = (
            ffmpeg.input(file, threads=0)
            .output("-", format="f32le", acodec="pcm_f32le", ac=1, ar=sr)
            .run(cmd=["ffmpeg", "-nostdin"], capture_stdout=True, capture_stderr=True)
        )
        return np.frombuffer(out, np.float32).flatten()
    except Exception as e:
        logger.debug(f"Failed to load audio with ffmpeg: {e}, trying librosa to load.")
    
    # Fallback 1: Try librosa (most common audio library)
    try:
        audio, orig_sr = librosa.load(file, sr=sr, mono=True)
        return audio.astype(np.float32)
    except Exception as e:
        logger.warning(f"Failed to load audio with librosa: {e}, trying soundfile...")
    
    # Fallback 2: Try soundfile
    try:
        audio, orig_sr = soundfile.read(file, dtype='float32')
        # Convert to mono if stereo
        if len(audio.shape) > 1:
            audio = audio.mean(axis=1)
        # Resample if necessary
        if orig_sr != sr:
            audio = librosa.resample(audio, orig_sr=orig_sr, target_sr=sr, res_type='kaiser_best')
        logger.info(f"Successfully loaded audio with soundfile (fallback method)")
        return audio.astype(np.float32)
    except Exception as e:
        logger.error(f"Failed to load audio with soundfile: {e}")
    
    # All methods failed
    raise RuntimeError(f"Failed to load audio file {file} with all available methods (ffmpeg, librosa, soundfile)")


def get_spepc(hps, filename):
    audio = load_audio(filename, int(hps.data.sampling_rate))
    audio = torch.FloatTensor(audio)
    audio_norm = audio
    audio_norm = audio_norm.unsqueeze(0)
    spec = spectrogram_torch(
        audio_norm,
        hps.data.filter_length,
        hps.data.sampling_rate,
        hps.data.hop_length,
        hps.data.win_length,
        center=False,
    )
    return spec


def get_language_texts(text):
    """
    Automatically detect and split text by language (zh, en, ja).
    Returns a list of dicts with 'lang' and 'text' keys.
    """
    # Define character ranges for different languages
    # Chinese: \u4e00-\u9fff
    # Japanese Hiragana: \u3040-\u309f
    # Japanese Katakana: \u30a0-\u30ff
    # English and numbers: a-zA-Z0-9
    pattern = re.compile(
        r'([\u4e00-\u9fff]+|[\u3040-\u309f\u30a0-\u30ff]+|[a-zA-Z0-9\s]+|[^\u4e00-\u9fff\u3040-\u309f\u30a0-\u30ffa-zA-Z0-9\s]+)'
    )
    
    results = []
    for match in pattern.finditer(text):
        segment = match.group(0)
        if not segment.strip():
            continue
            
        # Detect language
        if re.search(r'[\u4e00-\u9fff]', segment):
            lang = 'zh'
        elif re.search(r'[\u3040-\u309f\u30a0-\u30ff]', segment):
            lang = 'ja'
        elif re.search(r'[a-zA-Z]', segment):
            lang = 'en'
        else:
            # For punctuation and other characters, merge with previous segment
            if results:
                results[-1]['text'] += segment
            continue
        
        # Merge with previous segment if same language
        if results and results[-1]['lang'] == lang:
            results[-1]['text'] += segment
        else:
            results.append({'lang': lang, 'text': segment})
    
    return results


def split_en_inf(sentence, language):
    pattern = re.compile(r'[a-zA-Z ]+')
    textlist = []
    langlist = []
    pos = 0
    for match in pattern.finditer(sentence):
        start, end = match.span()
        if start > pos:
            textlist.append(sentence[pos:start])
            langlist.append(language)
        textlist.append(sentence[start:end])
        langlist.append("en")
        pos = end
    if pos < len(sentence):
        textlist.append(sentence[pos:])
        langlist.append(language)
    # Merge punctuation into previous word
    for i in range(len(textlist) - 1, 0, -1):
        if re.match(r'^[\W_]+$', textlist[i]):
            textlist[i - 1] += textlist[i]
            del textlist[i]
            del langlist[i]
    # Merge consecutive words with the same language tag
    i = 0
    while i < len(langlist) - 1:
        if langlist[i] == langlist[i + 1]:
            textlist[i] += textlist[i + 1]
            del textlist[i + 1]
            del langlist[i + 1]
        else:
            i += 1

    return textlist, langlist


def clean_text_inf(text, language):
    phones, word2ph, norm_text = clean_text(text, language)
    phones = cleaned_text_to_sequence(phones)
    return phones, word2ph, norm_text


class DictToAttrRecursive(dict):
    def __init__(self, input_dict):
        super().__init__(input_dict)
        for key, value in input_dict.items():
            if isinstance(value, dict):
                value = DictToAttrRecursive(value)
            self[key] = value
            setattr(self, key, value)

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(f"Attribute {item} not found")

    def __setattr__(self, key, value):
        if isinstance(value, dict):
            value = DictToAttrRecursive(value)
        super(DictToAttrRecursive, self).__setitem__(key, value)
        super().__setattr__(key, value)

    def __delattr__(self, item):
        try:
            del self[item]
        except KeyError:
            raise AttributeError(f"Attribute {item} not found")


def pcm16_header(rate, size=1000000000, channels=1):
    # Header for 16-bit PCM, modify from scipy.io.wavfile.write
    fs = rate
    # size = data.nbytes  # length * sizeof(nint16)

    header_data = b"RIFF"
    header_data += struct.pack("i", size + 44)
    header_data += b"WAVE"

    # fmt chunk
    header_data += b"fmt "
    format_tag = 1  # PCM
    bit_depth = 2 * 8  # 2 bytes
    bytes_per_second = fs * (bit_depth // 8) * channels
    block_align = channels * (bit_depth // 8)
    fmt_chunk_data = struct.pack("<HHIIHH", format_tag, channels, fs, bytes_per_second, block_align, bit_depth)

    header_data += struct.pack("<I", len(fmt_chunk_data))
    header_data += fmt_chunk_data
    header_data += b"data"
    return header_data + struct.pack("<I", size)


def save_wav(wav_data, filename, sample_rate=16000):
    with wave.open(filename, 'wb') as wf:
        wf.setnchannels(1)  # 单声道
        wf.setsampwidth(2)  # 16位样本，因此采样宽度为2字节
        wf.setframerate(sample_rate)  # 设置采样率
        wf.writeframes(wav_data)


def load_json_file(filepath):
    """
    加载指定路径的JSON文件，并返回其内容。
    """
    with open(filepath, 'r', encoding='utf-8') as file:
        data = json.load(file)
    return data


def merge_short_text_in_array(texts, threshold):
    if len(texts) < 2:
        return texts
    result = []
    text = ""
    for ele in texts:
        text += ele
        if len(text) >= threshold:
            result.append(text)
            text = ""
    if len(text) > 0:
        if len(result) == 0:
            result.append(text)
        else:
            result[len(result) - 1] += text
    return result


def get_first(text):
    pattern = "[" + "".join(re.escape(sep) for sep in sentence_split_symbols) + "]"
    text = re.split(pattern, text)[0].strip()
    return text


def sentence_split_by_symbol(input_text):
    input_text = input_text.replace("……", "。").replace("——", "，")
    if input_text[-1] not in sentence_split_symbols:
        input_text += "。"
    i_split_head = i_split_tail = 0
    len_text = len(input_text)
    split_sents = []
    while 1:
        if i_split_head >= len_text:
            break  # 结尾一定有标点，所以直接跳出即可，最后一段在上次已加入
        if input_text[i_split_head] in sentence_split_symbols:
            i_split_head += 1
            split_sents.append(input_text[i_split_tail:i_split_head])
            i_split_tail = i_split_head
        else:
            i_split_head += 1
    return split_sents


def sentence_split_by_length(input_text, max_len=50):
    """每50个字符切分一次"""
    input_text = input_text.strip("\n")
    sentences = sentence_split_by_symbol(input_text)
    if len(sentences) < 2:
        return input_text
    sents = []
    summ = 0
    tmp_str = ""
    for i in range(len(sentences)):
        summ += len(sentences[i])
        tmp_str += sentences[i]
        if summ > max_len:
            summ = 0
            sents.append(tmp_str)
            tmp_str = ""
    if tmp_str != "":
        sents.append(tmp_str)
    if len(sents) > 1 and len(sents[-1]) < max_len:  # 如果最后一个太短了，和前一个合一起
        sents[-2] = sents[-2] + sents[-1]
        sents = sents[:-1]
    return "\n".join(sents)


class TextToSpeech:
    def __init__(
            self,
            bert_model_path: str = "shibing624/parrots-chinese-roberta-wwm-ext-large",
            hubert_model_path: str = "shibing624/parrots-chinese-hubert-base",
            sovits_model_path: str = None,
            gpt_model_path: str = None,
            speaker_model_path: str = "shibing624/parrots-gpt-sovits-speaker-maimai",
            speaker_name: Optional[str] = "MaiMai",
            device: Optional[str] = None,
            half: Optional[bool] = False,
            ensure_nltk: bool = True,
    ):
        """
        Args:
            bert_model_path: str, path to the pretrained BERT model
            hubert_model_path: str, path to the pretrained HuBERT model
            sovits_model_path: str, path to the pretrained SoVITS, if None, use the speaker_model_path
            gpt_model_path: str, path to the pretrained GPT, if None, use the speaker_model_path
            speaker_model_path: str, path to the pretrained speaker model,
                if sovits_model_path and gpt_model_path are None, use this model, else ignore it
            speaker_name: str, name of the speaker, default is "MaiMai"
            device: str, device to run on, "cuda", "cpu" or "mps"
            half: bool, use half precision instead of float32
            ensure_nltk: bool, whether to check/download NLTK resources during initialization
        """
        if device is None:
            if torch.cuda.is_available():
                device = "cuda"
            elif torch.backends.mps.is_available():
                device = "cpu"
            else:
                device = "cpu"
        self.device = device
        logger.debug("Use device: {}".format(self.device))
        self.half = half
        self.dtype = torch.float16 if half else torch.float32

        # Ensure required NLTK resources are available (optional to avoid import-time side effects)
        if ensure_nltk:
            self._ensure_nltk_data()

        # BERT
        self.bert_tokenizer = AutoTokenizer.from_pretrained(bert_model_path)
        self.bert_model = AutoModelForMaskedLM.from_pretrained(bert_model_path)
        self.bert_model = self.to_device(self.bert_model)

        # Speaker
        if bool(sovits_model_path) != bool(gpt_model_path):  # Check if only one of them is provided
            raise ValueError("sovits_model_path and gpt_model_path must be provided together")
        if sovits_model_path is None and gpt_model_path is None:
            if speaker_model_path:
                logger.info("Load pretrained parrots speaker: {}".format(speaker_model_path))
                if os.path.exists(speaker_model_path):
                    # Load from path
                    model_path = speaker_model_path
                else:
                    # Load from huggingface model hub
                    model_path = snapshot_download(speaker_model_path)
                sovits_model_path = os.path.join(model_path, speaker_name, SOVITS_MODEL_NAME)
                gpt_model_path = os.path.join(model_path, speaker_name, GPT_MODEL_NAME)
                # Set ref_wav_path to the speaker's reference wav file
                config_file = os.path.join(model_path, speaker_name, CONFIG_NAME)
                if os.path.exists(config_file):
                    ref_config = load_json_file(config_file)
                    logger.debug(f"Reference speaker config: {ref_config}, loaded from {config_file}")
                    self.ref_wav_path = os.path.join(model_path, speaker_name, REF_WAV_NAME)
                    self.ref_prompt = ref_config.get("reference_prompt", "")
                    self.ref_language = ref_config.get("reference_language", "")
                else:
                    raise ValueError(f"Config file not found: {config_file}")
            else:
                raise ValueError("sovits_model_path, gpt_model_path or speaker_model_path must be provided")

        # SoVITS
        sovits_dict = torch.load(sovits_model_path, map_location="cpu", weights_only=False)
        hps = DictToAttrRecursive(sovits_dict["config"])
        logger.debug(f"SoVITS config: {hps}")
        vq_model = SynthesizerModel(
            spec_channels=hps.data.filter_length // 2 + 1,
            segment_size=hps.train.segment_size // hps.data.hop_length,
            n_speakers=hps.data.n_speakers,
            **hps.model,
        )
        self.sampling_rate = hps.data.sampling_rate
        self.hps = hps
        vq_model = self.to_device(vq_model)
        vq_model.eval()
        vq_model.load_state_dict(sovits_dict["weight"], strict=False)
        self.vq_model = vq_model

        # GPT
        gpt_dict = torch.load(gpt_model_path, map_location="cpu", weights_only=False)
        config = gpt_dict["config"]
        logger.debug(f"GPT config: {config}")
        t2s_model = Text2SemanticDecoder(config)
        # Convert Text2SemanticLightningModule to Text2SemanticDecoder model
        adjusted_state_dict = self.adjust_keys(gpt_dict["weight"])
        t2s_model.load_state_dict(adjusted_state_dict)
        t2s_model = self.to_device(t2s_model)
        t2s_model.eval()
        self.t2s_model = t2s_model
        self.t2s_max_sec = config["data"]["max_sec"]
        total = sum([param.nelement() for param in t2s_model.parameters()])
        logger.debug("Number of t2s model parameter: %.2fM" % (total / 1e6))

        # HuBERT
        self.ssl_model = self.to_device(cnhubert.get_model(hubert_model_path, self.sampling_rate))
        
        # Cache for streaming TTS
        self._ref_cache = {}

    def _ensure_nltk_data(self):
        """Ensure NLTK tagger resources exist, download quietly if missing.
        Moved from module scope to avoid side effects at import time for PyPI.
        """
        try:
            try:
                nltk.data.find('taggers/averaged_perceptron_tagger_eng')
            except LookupError:
                logger.info("Downloading NLTK averaged_perceptron_tagger_eng...")
                nltk.download('averaged_perceptron_tagger_eng', quiet=True)
            try:
                nltk.data.find('taggers/averaged_perceptron_tagger')
            except LookupError:
                logger.info("Downloading NLTK averaged_perceptron_tagger...")
                nltk.download('averaged_perceptron_tagger', quiet=True)
        except Exception as e:
            logger.warning(f"Failed to ensure NLTK data: {e}")

    @staticmethod
    def adjust_keys(state_dict):
        new_state_dict = {}
        # 对于每一个在状态字典中的键和值
        for k, v in state_dict.items():
            # 如果键以"model."开始，就去掉这个前缀
            if k.startswith("model."):
                new_key = k[len("model."):]  # 删除前缀
            else:
                new_key = k
            new_state_dict[new_key] = v  # 加入调整后的键值对到新字典
        return new_state_dict

    def to_device(self, model):
        """A helper function to move a model to GPU or cpu"""
        if self.half:
            model = model.half()
        return model.to(self.device)

    def get_bert_feature(self, text, word2ph):
        with torch.no_grad():
            inputs = self.bert_tokenizer(text, return_tensors="pt")
            for i in inputs:
                inputs[i] = inputs[i].to(self.bert_model.device)
            res = self.bert_model(**inputs, output_hidden_states=True)
            res = torch.cat(res["hidden_states"][-3:-2], -1)[0].cpu()[1:-1]
        assert len(word2ph) == len(text)
        phone_level_feature = []
        for i in range(len(word2ph)):
            repeat_feature = res[i].repeat(word2ph[i], 1)
            phone_level_feature.append(repeat_feature)
        phone_level_feature = torch.cat(phone_level_feature, dim=0)
        return phone_level_feature.T

    def get_bert_inf(self, phones, word2ph, norm_text, language):
        if language == "zh":
            bert = self.get_bert_feature(norm_text, word2ph).to(self.device)
        else:
            bert = torch.zeros(
                (1024, len(phones)), dtype=torch.float16 if self.half else torch.float32,
            ).to(self.device)
        return bert

    def nonen_get_bert_inf(self, text, language):
        if language == "auto":
            textlist = []
            langlist = []
            for tmp in get_language_texts(text):
                langlist.append(tmp["lang"])
                textlist.append(tmp["text"])
        else:
            textlist, langlist = split_en_inf(text, language)
        bert_list = []
        for i in range(len(textlist)):
            lang = langlist[i]
            phones, word2ph, norm_text = clean_text_inf(textlist[i], lang)
            bert_inf = self.get_bert_inf(phones, word2ph, norm_text, lang)
            bert_list.append(bert_inf)
        return torch.cat(bert_list, dim=1)

    def nonen_clean_text_inf(self, text, language):
        if language == "auto":
            textlist = []
            langlist = []
            for tmp in get_language_texts(text):
                langlist.append(tmp["lang"])
                textlist.append(tmp["text"])
        else:
            textlist, langlist = split_en_inf(text, language)
        logger.debug(textlist)
        logger.debug(langlist)
        phones_list = []
        word2ph_list = []
        norm_text_list = []
        for i in range(len(textlist)):
            lang = langlist[i]
            phones, word2ph, norm_text = clean_text_inf(textlist[i], lang)
            phones_list.append(phones)
            if lang == "zh":
                word2ph_list.append(word2ph)
            norm_text_list.append(norm_text)
        phones = sum(phones_list, [])
        word2ph = sum(word2ph_list, [])
        norm_text = ' '.join(norm_text_list)
        return phones, word2ph, norm_text

    def get_cleaned_text_final(self, text, language):
        if language in {"en", "zh", "ja"}:
            phones, word2ph, norm_text = clean_text_inf(text, language)
        elif language in {"auto"}:
            phones, word2ph, norm_text = self.nonen_clean_text_inf(text, language)
        else:
            raise ValueError(f"Unsupported language: {language}")
        return phones, word2ph, norm_text

    def get_bert_final(self, phones, word2ph, text, language):
        if language == "en":
            bert = self.get_bert_inf(phones, word2ph, text, language)
        elif language in {"ja", "auto"}:
            bert = self.nonen_get_bert_inf(text, language)
        elif language == "zh":
            bert = self.get_bert_feature(text, word2ph).to(self.device)
        else:
            bert = torch.zeros((1024, len(phones))).to(self.device)
        return bert

    @staticmethod
    def audio_numpy_concat(segment_data_list, sr, speed=1.):
        audio_segments = []
        for segment_data in segment_data_list:
            audio_segments += segment_data.reshape(-1).tolist()
            audio_segments += [0] * int((sr * 0.05) / speed)
        audio_segments = np.array(audio_segments).astype(np.float32)
        return audio_segments

    @torch.inference_mode()
    def predict(
            self,
            text: str,
            text_language: Union[str, LANG] = "auto",
            speed: float = 1.0,
            output_path: Optional[str] = None,
            ref_wav_path: str = None,
            ref_prompt: str = None,
            ref_language: Union[str, LANG] = None,
            top_k: int = 20,
            top_p: float = 0.6,
            temperature: float = 0.6,
    ):
        """
        Args:
            text: str, target text 要语音合成的文本
            text_language: str, language of the target text 要语音合成的文本的语种
            speed: float, speed of speech 语速
            output_path: str, path to save the output wav file 保存语音合成结果的路径，可选
            ref_wav_path: str, path to the reference wav file 参考音频文件
            ref_prompt: str, reference prompt 参考音频对应的文本
            ref_language: str, language of the reference prompt 参考音频对应的文本的语种
            top_k: int, top k
            top_p: float, top p
            temperature: float, temperature
        Returns:
            audio array: generator, audio stream, numpy array
        """
        if ref_wav_path is None:
            ref_wav_path = self.ref_wav_path
        if ref_prompt is None:
            ref_prompt = self.ref_prompt
        if ref_language is None:
            ref_language = self.ref_language

        ref_language = LANG.from_string(ref_language) if isinstance(ref_language, str) else ref_language
        if ref_language not in list(LANG):
            raise ValueError(f"ref_language must be in {list(LANG)}")
        ref_language = str(ref_language)
        text_language = LANG.from_string(text_language) if isinstance(text_language, str) else text_language
        if text_language not in list(LANG):
            raise ValueError(f"text_language must be in {list(LANG)}")
        text_language = str(text_language)

        ref_prompt = ref_prompt.strip()
        text = text.strip()
        if ref_prompt[-1] not in sentence_split_symbols:
            ref_prompt += "。" if ref_language != "en" else "."
        if text[0] not in sentence_split_symbols and len(get_first(text)) < 4:
            text = "。" + text if text_language != "en" else "." + text

        zero_wav = np.zeros(
            int(self.hps.data.sampling_rate * 0.3),
            dtype=np.float16 if self.half else np.float32,
        )
        wav16k, sr = librosa.load(ref_wav_path, sr=16000)
        if wav16k.shape[0] > 160000 or wav16k.shape[0] < 48000:
            raise ValueError("参考音频需要是3~10秒范围内的，请更换！")
        wav16k = torch.from_numpy(wav16k)
        zero_wav_torch = torch.from_numpy(zero_wav)
        if self.half:
            wav16k = wav16k.half().to(self.device)
            zero_wav_torch = zero_wav_torch.half().to(self.device)
        else:
            wav16k = wav16k.to(self.device)
            zero_wav_torch = zero_wav_torch.to(self.device)
        wav16k = torch.cat([wav16k, zero_wav_torch])

        ssl_content = self.ssl_model.model(wav16k.unsqueeze(0))["last_hidden_state"].transpose(1, 2)
        codes = self.vq_model.extract_latent(ssl_content)
        prompt_semantic = codes[0, 0]

        # step1
        phones1, word2ph1, norm_text1 = self.get_cleaned_text_final(ref_prompt, ref_language)
        text = sentence_split_by_length(text)

        while "\n\n" in text:
            text = text.replace("\n\n", "\n")
        texts = text.split("\n")
        texts = merge_short_text_in_array(texts, 5)

        audio_list = []
        bert1 = self.get_bert_final(phones1, word2ph1, norm_text1, ref_language).to(self.dtype)

        for text in texts:
            if not text.strip():
                continue
            if text[-1] not in sentence_split_symbols:
                text += "。" if text_language != "en" else "."
            logger.debug(f"实际输入的目标文本(每句): {text}")
            phones2, word2ph2, norm_text2 = self.get_cleaned_text_final(text, text_language)
            bert2 = self.get_bert_final(phones2, word2ph2, norm_text2, text_language).to(self.dtype)
            bert = torch.cat([bert1, bert2], 1)

            all_phoneme_ids = torch.LongTensor(phones1 + phones2).to(self.device).unsqueeze(0)
            bert = bert.to(self.device).unsqueeze(0)
            all_phoneme_len = torch.tensor([all_phoneme_ids.shape[-1]]).to(self.device)
            prompt = prompt_semantic.unsqueeze(0).to(self.device)

            # step2
            pred_semantic, idx = self.t2s_model.infer_panel(
                all_phoneme_ids,
                all_phoneme_len,
                prompt,
                bert,
                top_k=top_k,
                top_p=top_p,
                temperature=temperature,
                early_stop_num=50 * self.t2s_max_sec,
            )

            # step3
            pred_semantic = pred_semantic[:, -idx:].unsqueeze(0)
            refer = get_spepc(self.hps, ref_wav_path)
            refer = self.to_device(refer)
            # audio = vq_model.decode(pred_semantic, all_phoneme_ids, refer).detach().cpu().numpy()[0, 0]
            audio = (
                self.vq_model.decode(
                    pred_semantic,
                    torch.LongTensor(phones2).to(self.device).unsqueeze(0),
                    refer,
                ).detach().cpu().numpy()[0, 0]
            )  # 重建不带上prompt部分
            max_audio = np.abs(audio).max()  # 简单防止16bit爆音
            if max_audio > 1:
                audio /= max_audio
            audio_list.append(audio)
        audio_output = self.audio_numpy_concat(audio_list, sr=self.sampling_rate, speed=speed)
        if output_path is None:
            return audio_output
        else:
            soundfile.write(output_path, audio_output, self.sampling_rate)
            logger.debug(f"Save audio to {output_path}")
            return output_path

    @torch.inference_mode()
    def prepare_reference(
            self,
            ref_wav_path: str = None,
            ref_prompt: str = None,
            ref_language: Union[str, LANG] = None,
    ):
        """
        Prepare and cache reference audio features for streaming TTS.
        This significantly reduces first chunk latency by pre-computing reference features.
        
        Args:
            ref_wav_path: str, path to the reference wav file
            ref_prompt: str, reference prompt text
            ref_language: str, language of the reference prompt
            
        Returns:
            cache_key: str, key for accessing cached features
        """
        if ref_wav_path is None:
            ref_wav_path = self.ref_wav_path
        if ref_prompt is None:
            ref_prompt = self.ref_prompt
        if ref_language is None:
            ref_language = self.ref_language
            
        ref_language = LANG.from_string(ref_language) if isinstance(ref_language, str) else ref_language
        ref_language = str(ref_language)
        
        ref_prompt = ref_prompt.strip()
        if ref_prompt[-1] not in sentence_split_symbols:
            ref_prompt += "。" if ref_language != "en" else "."
            
        # Create cache key
        cache_key = f"{ref_wav_path}_{ref_prompt}_{ref_language}"
        
        # Check if already cached
        if cache_key in self._ref_cache:
            logger.debug(f"Reference already cached: {cache_key}")
            return cache_key
            
        logger.info(f"Preparing reference audio features for streaming...")
        
        # Load and process reference audio
        zero_wav = np.zeros(
            int(self.hps.data.sampling_rate * 0.3),
            dtype=np.float16 if self.half else np.float32,
        )
        wav16k, sr = librosa.load(ref_wav_path, sr=16000)
        if wav16k.shape[0] > 160000 or wav16k.shape[0] < 48000:
            raise ValueError("参考音频需要是3~10秒范围内的，请更换！")
            
        wav16k = torch.from_numpy(wav16k)
        zero_wav_torch = torch.from_numpy(zero_wav)
        if self.half:
            wav16k = wav16k.half().to(self.device)
            zero_wav_torch = zero_wav_torch.half().to(self.device)
        else:
            wav16k = wav16k.to(self.device)
            zero_wav_torch = zero_wav_torch.to(self.device)
        wav16k = torch.cat([wav16k, zero_wav_torch])
        
        # Extract semantic features
        ssl_content = self.ssl_model.model(wav16k.unsqueeze(0))["last_hidden_state"].transpose(1, 2)
        codes = self.vq_model.extract_latent(ssl_content)
        prompt_semantic = codes[0, 0]
        
        # Process reference text
        phones1, word2ph1, norm_text1 = self.get_cleaned_text_final(ref_prompt, ref_language)
        bert1 = self.get_bert_final(phones1, word2ph1, norm_text1, ref_language).to(self.dtype)
        
        # Get reference spectrogram
        refer = get_spepc(self.hps, ref_wav_path)
        refer = self.to_device(refer)
        
        # Cache all features
        self._ref_cache[cache_key] = {
            "prompt_semantic": prompt_semantic,
            "phones1": phones1,
            "word2ph1": word2ph1,
            "norm_text1": norm_text1,
            "bert1": bert1,
            "refer": refer,
            "ref_language": ref_language,
        }
        
        logger.info(f"Reference prepared and cached: {cache_key}")
        return cache_key

    @torch.inference_mode()
    def predict_stream(
            self,
            text: str,
            text_language: Union[str, LANG] = "auto",
            speed: float = 1.0,
            ref_wav_path: str = None,
            ref_prompt: str = None,
            ref_language: Union[str, LANG] = None,
            top_k: int = 20,
            top_p: float = 0.6,
            temperature: float = 0.6,
            stream_chunk_size: int = 20,
    ):
        """
        Streaming TTS generation with incremental decoding.
        Accumulates all semantic tokens and decodes incrementally, outputting only new audio.
        
        Args:
            text: str, target text
            text_language: str, language of target text
            speed: float, speech speed
            ref_wav_path: str, reference audio file
            ref_prompt: str, reference prompt text
            ref_language: str, reference language
            top_k: int, top k sampling
            top_p: float, top p sampling
            temperature: float, sampling temperature
            stream_chunk_size: int, number of semantic tokens per chunk (smaller = lower latency)
            
        Yields:
            audio_chunk: numpy array, incremental audio chunk (float32)
        """
        # Prepare reference (use cache if available)
        cache_key = self.prepare_reference(ref_wav_path, ref_prompt, ref_language)
        ref_cache = self._ref_cache[cache_key]
        
        # Parse language
        text_language = LANG.from_string(text_language) if isinstance(text_language, str) else text_language
        text_language = str(text_language)
        
        # Preprocess text
        text = text.strip()
        if text[0] not in sentence_split_symbols and len(get_first(text)) < 4:
            text = "。" + text if text_language != "en" else "." + text
            
        text = sentence_split_by_length(text)
        while "\n\n" in text:
            text = text.replace("\n\n", "\n")
        texts = text.split("\n")
        texts = merge_short_text_in_array(texts, 5)
        
        logger.debug(f"Streaming TTS for {len(texts)} text segments")
        
        # Process each text segment
        for text_idx, text in enumerate(texts):
            if not text.strip():
                continue
                
            if text[-1] not in sentence_split_symbols:
                text += "。" if text_language != "en" else "."
                
            logger.debug(f"Processing segment {text_idx + 1}/{len(texts)}: {text}")
            
            # Text processing
            phones2, word2ph2, norm_text2 = self.get_cleaned_text_final(text, text_language)
            bert2 = self.get_bert_final(phones2, word2ph2, norm_text2, text_language).to(self.dtype)
            bert = torch.cat([ref_cache["bert1"], bert2], 1)
            
            all_phoneme_ids = torch.LongTensor(ref_cache["phones1"] + phones2).to(self.device).unsqueeze(0)
            bert = bert.to(self.device).unsqueeze(0)
            all_phoneme_len = torch.tensor([all_phoneme_ids.shape[-1]]).to(self.device)
            prompt = ref_cache["prompt_semantic"].unsqueeze(0).to(self.device)
            
            # Streaming with chunk-by-chunk decoding
            # Strategy: decode each semantic chunk independently to avoid alignment issues
            for semantic_chunk in self._generate_semantic_stream(
                all_phoneme_ids,
                all_phoneme_len,
                prompt,
                bert,
                top_k=top_k,
                top_p=top_p,
                temperature=temperature,
                chunk_size=stream_chunk_size,
            ):
                try:
                    # Decode only the current semantic chunk
                    # This avoids alignment issues between accumulated semantics and phonemes
                    audio_chunk = self.vq_model.decode(
                        semantic_chunk.unsqueeze(0),
                        torch.LongTensor(phones2).to(self.device).unsqueeze(0),
                        ref_cache["refer"],
                    ).detach().cpu().numpy()[0, 0]
                    
                    # Normalize audio
                    max_audio = np.abs(audio_chunk).max()
                    if max_audio > 1:
                        audio_chunk /= max_audio
                    
                    # Apply speed adjustment if needed
                    if speed != 1.0:
                        audio_chunk = librosa.effects.time_stretch(audio_chunk, rate=speed, sr=self.sampling_rate)
                    
                    # Yield the audio chunk
                    if len(audio_chunk) > 0:  # Only yield non-empty chunks
                        yield audio_chunk.astype(np.float32)
                    
                except Exception as e:
                    logger.warning(f"Failed to decode chunk: {e}")
                    continue

    def _generate_semantic_stream(
            self,
            all_phoneme_ids,
            all_phoneme_len,
            prompt,
            bert,
            top_k=20,
            top_p=0.6,
            temperature=0.6,
            chunk_size=20,
    ):
        """
        Generate semantic tokens in streaming fashion.
        Yields chunks of semantic tokens as they are generated.
        (Legacy method - kept for compatibility)
        """
        from parrots.t2s_utils import sample
        import torch.nn.functional as F
        
        # Prepare text embeddings
        x = self.t2s_model.ar_text_embedding(all_phoneme_ids)
        x = x + self.t2s_model.bert_proj(bert.transpose(1, 2))
        x = self.t2s_model.ar_text_position(x)
        
        y = prompt
        prefix_len = y.shape[1]
        x_len = x.shape[1]
        x_attn_mask = torch.zeros((x_len, x_len), dtype=torch.bool, device=x.device)
        
        # Cache for efficient generation
        cache = {
            "all_stage": self.t2s_model.num_layers,
            "k": [None] * self.t2s_model.num_layers,
            "v": [None] * self.t2s_model.num_layers,
            "y_emb": None,
            "first_infer": 1,
            "stage": 0,
        }
        
        chunk_tokens = []
        
        for idx in range(1500):  # Max length
            # Generate next token
            if cache["first_infer"] == 1:
                y_emb = self.t2s_model.ar_audio_embedding(y)
            else:
                y_emb = torch.cat([cache["y_emb"], self.t2s_model.ar_audio_embedding(y[:, -1:])], 1)
            cache["y_emb"] = y_emb
            
            y_pos = self.t2s_model.ar_audio_position(y_emb)
            
            if cache["first_infer"] == 1:
                xy_pos = torch.concat([x, y_pos], dim=1)
            else:
                xy_pos = y_pos[:, -1:]
                
            y_len = y_pos.shape[1]
            
            if cache["first_infer"] == 1:
                x_attn_mask_pad = F.pad(x_attn_mask, (0, y_len), value=True)
                y_attn_mask = F.pad(
                    torch.triu(torch.ones(y_len, y_len, dtype=torch.bool, device=y.device), diagonal=1),
                    (x_len, 0),
                    value=False,
                )
                xy_attn_mask = torch.concat([x_attn_mask_pad, y_attn_mask], dim=0)
            else:
                xy_attn_mask = torch.zeros((1, x_len + y_len), dtype=torch.bool, device=xy_pos.device)
                
            xy_dec, _ = self.t2s_model.h((xy_pos, None), mask=xy_attn_mask, cache=cache)
            logits = self.t2s_model.ar_predict_layer(xy_dec[:, -1])
            
            if idx == 0:
                logits = logits[:, :-1]
                
            samples = sample(
                logits[0], y, top_k=top_k, top_p=top_p, repetition_penalty=1.35, temperature=temperature
            )[0].unsqueeze(0)
            
            # Check for EOS
            if torch.argmax(logits, dim=-1)[0] == self.t2s_model.EOS or samples[0, 0] == self.t2s_model.EOS:
                if len(chunk_tokens) > 0:
                    yield torch.cat(chunk_tokens, dim=1)
                break
                
            y = torch.concat([y, samples], dim=1)
            cache["first_infer"] = 0
            
            # Accumulate tokens for chunk
            if y.shape[1] > prefix_len:
                chunk_tokens.append(y[:, -1:])
                
                # Yield chunk when reaching chunk_size
                if len(chunk_tokens) >= chunk_size:
                    yield torch.cat(chunk_tokens, dim=1)
                    chunk_tokens = []


if __name__ == "__main__":
    sys.path.append('..')
    parser = argparse.ArgumentParser(description="TTS")
    parser.add_argument(
        "--bert",
        type=str,
        default="pretrained_models/chinese-roberta-wwm-ext-large",
        help="Path to the pretrained BERT model",
    )
    parser.add_argument(
        "--hubert",
        type=str,
        default="pretrained_models/chinese-hubert-base",
        help="Path to the pretrained HuBERT model",
    )
    parser.add_argument(
        "--sovits",
        type=str,
        default="pretrained_models/s2G488k.pth",
        help="Path to the pretrained SoVITS",
    )
    parser.add_argument(
        "--gpt",
        type=str,
        default="pretrained_models/s1bert25hz-2kh-longer-epoch=68e-step=50232.ckpt",
        help="Path to the pretrained GPT",
    )
    parser.add_argument("--device", type=str, default="cuda", help="Device to run on")
    parser.add_argument("--half", action="store_true", help="Use half precision instead of float32")
    parser.add_argument("--text", type=str, default="你好，欢迎来北京。welcome to the city.", help="input text")
    parser.add_argument("--lang", type=str, default="auto", help="Language of the text, zh, en, jp, auto")
    parser.add_argument("--ref_wav_path", type=str, default="../examples/ref.wav", help="reference wav")
    parser.add_argument("--ref_text", type=str,
                        default="大家好，我是宁宁。我中文还不是很熟练，但是希望大家能喜欢我的声音，喵喵喵！",
                        help="reference text")
    parser.add_argument("--ref_lang", type=str, default="zh", help="reference wav language")
    parser.add_argument("--output_path", type=str, default="output_audio.wav", help="output wav")
    args = parser.parse_args()
    print(f"args: {args}")
    m = TextToSpeech(
        bert_model_path=args.bert,
        hubert_model_path=args.hubert,
        sovits_model_path=args.sovits,
        gpt_model_path=args.gpt,
        device=args.device,
        half=args.half
    )
    m.predict(
        ref_wav_path=args.ref_wav_path,
        ref_prompt=args.ref_text,
        ref_language=args.ref_lang,
        text=args.text,
        text_language=args.lang,
        output_path=args.output_path
    )



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""
import math
import wave
import numpy as np
import difflib


def edit_distance(str1, str2):
    distance = 0
    s = difflib.SequenceMatcher(None, str1, str2)
    for tag, i1, i2, j1, j2 in s.get_opcodes():
        # print('{:7} a[{}: {}] --> b[{}: {}] {} --> {}'.format(tag, i1, i2, j1, j2, str1[i1: i2], str2[j1: j2]))
        if tag == 'replace':
            distance += max(i2 - i1, j2 - j1)
        elif tag == 'insert':
            distance += (j2 - j1)
        elif tag == 'delete':
            distance += (i2 - i1)
    return distance


def read_wav_data(filename):
    """
    读取一个wav文件，返回声音信号的时域谱矩阵和播放时间
    """
    wav = wave.open(filename, "rb")  # 打开一个wav格式的声音文件流
    num_frame = wav.getnframes()  # 获取帧数
    num_channel = wav.getnchannels()  # 获取声道数
    framerate = wav.getframerate()  # 获取帧速率
    str_data = wav.readframes(num_frame)  # 读取全部的帧
    wav.close()  # 关闭流
    wave_data = np.fromstring(str_data, dtype=np.short)  # 将声音文件数据转换为数组矩阵形式
    wave_data.shape = -1, num_channel  # 按照声道数将数组整形，单声道时候是一列数组，双声道时候是两列的矩阵
    wave_data = wave_data.T  # 将矩阵转置
    return wave_data, framerate


def GetFrequencyFeature(wavsignal, fs):
    from scipy.fftpack import fft
    # wav波形 加时间窗以及时移10ms
    time_window = 25  # 单位ms
    data_input = []

    # print(int(len(wavsignal[0])/fs*1000 - time_window) // 10)
    wav_length = len(wavsignal[0])  # 计算一条语音信号的原始长度
    range0_end = int(len(wavsignal[0]) / fs * 1000 - time_window) // 10  # 计算循环终止的位置，也就是最终生成的窗数
    for i in range(0, range0_end):
        p_start = i * 160
        p_end = p_start + 400
        data_line = []

        for j in range(p_start, p_end):
            data_line.append(wavsignal[0][j])
        # print('wavsignal[0][j]:\n',wavsignal[0][j])
        # data_line = abs(fft(data_line)) / len(wavsignal[0])
        data_line = fft(data_line) / wav_length
        data_line2 = []
        for fre_sig in data_line:
            # 分别取出频率信号的实部和虚部作为语音信号的频率特征
            # 直接使用复数的话，之后会被numpy将虚部丢弃，造成信息丢失
            # print('fre_sig:\n',fre_sig)
            data_line2.append(fre_sig.real)
            data_line2.append(fre_sig.imag)

        data_input.append(data_line2[0:len(data_line2) // 2])  # 除以2是取一半数据，因为是对称的
    return data_input


def GetFrequencyFeature2(wavsignal, fs):
    from scipy.fftpack import fft
    # wav波形 加时间窗以及时移10ms
    time_window = 25  # 单位ms
    window_length = fs / 1000 * time_window  # 计算窗长度的公式，目前全部为400固定值

    wav_arr = np.array(wavsignal)
    # wav_length = len(wavsignal[0])
    wav_length = wav_arr.shape[1]

    range0_end = int(len(wavsignal[0]) / fs * 1000 - time_window) // 10  # 计算循环终止的位置，也就是最终生成的窗数
    data_input = np.zeros((range0_end, 200), dtype=np.float)  # 用于存放最终的频率特征数据
    data_line = np.zeros((1, 400), dtype=np.float)
    for i in range(0, range0_end):
        p_start = i * 160
        p_end = p_start + 400

        data_line = wav_arr[0, p_start:p_end]
        """
        x=np.linspace(0, 400 - 1, 400, dtype = np.int64)
        w = 0.54 - 0.46 * np.cos(2 * np.pi * (x) / (400 - 1) ) # 汉明窗
        data_line = data_line * w # 加窗
        """
        data_line = np.abs(fft(data_line)) / wav_length

        data_input[i] = data_line[0:200]  # 设置为400除以2的值(即200）是取一半数据，因为是对称的

    # print(data_input.shape)
    return data_input


x = np.linspace(0, 400 - 1, 400, dtype=np.int64)
w = 0.54 - 0.46 * np.cos(2 * np.pi * (x) / (400 - 1))  # 汉明窗


def get_frequency_features(wavsignal, fs):
    from scipy.fftpack import fft
    # wav波形 加时间窗以及时移10ms
    time_window = 25  # 单位ms
    window_length = fs / 1000 * time_window  # 计算窗长度的公式，目前全部为400固定值

    wav_arr = np.array(wavsignal)
    wav_length = wav_arr.shape[1]

    range0_end = int(len(wavsignal[0]) / fs * 1000 - time_window) // 10  # 计算循环终止的位置，也就是最终生成的窗数
    data_input = np.zeros((range0_end, 200), dtype=np.float)  # 用于存放最终的频率特征数据
    data_line = np.zeros((1, 400), dtype=np.float)

    for i in range(0, range0_end):
        p_start = i * 160
        p_end = p_start + 400
        data_line = wav_arr[0, p_start:p_end]
        data_line = data_line * w  # 加窗
        data_line = np.abs(fft(data_line)) / wav_length
        data_input[i] = data_line[0:200]  # 设置为400除以2的值(即200）是取一半数据，因为是对称的

    # print(data_input.shape)
    data_input = np.log(data_input + 1)
    return data_input


def wav_scale(energy):
    """
    语音信号能量归一化
    """
    means = energy.mean()  # 均值
    var = energy.var()  # 方差
    e = (energy - means) / math.sqrt(var)  # 归一化能量
    return e


def wav_scale2(energy):
    """
    语音信号能量归一化
    """
    maxnum = max(energy)
    e = energy / maxnum
    return e


def wav_scale3(energy):
    """
    语音信号能量归一化
    """
    for i in range(len(energy)):
        # if i == 1:
        #	#print('wavsignal[0]:\n {:.4f}'.format(energy[1]),energy[1] is int)
        energy[i] = float(energy[i]) / 100.0
    # if i == 1:
    #	#print('wavsignal[0]:\n {:.4f}'.format(energy[1]),energy[1] is int)
    return energy


def get_wav_list(filename):
    """
    读取一个wav文件列表，返回一个存储该列表的字典类型值
    ps:在数据中专门有几个文件用于存放用于训练、验证和测试的wav文件列表
    """
    txt_obj = open(filename, 'r')  # 打开文件并读入
    txt_text = txt_obj.read()
    txt_lines = txt_text.split('\n')  # 文本分割
    dic_filelist = {}  # 初始化字典
    list_wavmark = []  # 初始化wav列表
    for i in txt_lines:
        if i != '':
            txt_l = i.split(' ')
            dic_filelist[txt_l[0]] = txt_l[1]
            list_wavmark.append(txt_l[0])
    txt_obj.close()
    return dic_filelist, list_wavmark


def get_wav_symbol(filename):
    """
    读取指定数据集中，所有wav文件对应的语音符号
    返回一个存储符号集的字典类型值
    """
    txt_obj = open(filename, 'r')  # 打开文件并读入
    txt_text = txt_obj.read()
    txt_lines = txt_text.split('\n')  # 文本分割
    dic_symbol_list = {}  # 初始化字典
    list_symbolmark = []  # 初始化symbol列表
    for i in txt_lines:
        if i != '':
            txt_l = i.split(' ')
            dic_symbol_list[txt_l[0]] = txt_l[1:]
            list_symbolmark.append(txt_l[0])
    txt_obj.close()
    return dic_symbol_list, list_symbolmark


def get_pinyin_list(dict_path=''):
    """
    加载拼音符号列表，用于标记符号
    :param dict_path: 拼音符号列表
    :return:
    """
    list_symbol = []  # 符号列表
    pinyin_idx = 0
    with open(dict_path, mode='r', encoding='UTF-8') as f:
        for line in f:
            line = line.strip('\n')
            parts = line.split('\t')
            list_symbol.append(parts[pinyin_idx])
    list_symbol.append('_')
    return list_symbol



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Unified logger configuration for parrots project
"""
import os
import sys
from loguru import logger  as _logger

# Remove default logger
_logger.remove()

# Default log level
LOG_LEVEL = os.environ.get("PARROTS_LOG_LEVEL", "INFO").upper()



# Add custom logger with configurable level
_logger.add(
    sys.stderr,
    format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>",
    level=LOG_LEVEL,
    colorize=True,
)


def set_log_level(level: str):
    """
    Set the log level for the parrots logger.
    
    Args:
        level: str, one of "TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"
    
    Examples:
        >>> from parrots.log import set_log_level
        >>> set_log_level("DEBUG")  # Enable debug logging
        >>> set_log_level("WARNING")  # Only show warnings and errors
    """
    level = level.upper()
    valid_levels = ["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"]
    
    if level not in valid_levels:
        raise ValueError(f"Invalid log level: {level}. Must be one of {valid_levels}")
    
    # Remove all handlers and add new one with updated level
    _logger.remove()
    _logger.add(
        sys.stderr,
        format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>",
        level=level,
        colorize=True,
    )


def add_file_logger(log_file: str, level: str = "DEBUG"):
    """
    Add a file handler to the logger.

    Args:
        log_file: Path to the log file
        level: Log level for the file handler (default: DEBUG)
    """
    _logger.add(
        log_file,
        format="{time:YYYY-MM-DD HH:mm:ss} | {level: <8} | {name}:{function}:{line} - {message}",
        level=level.upper(),
        rotation="10 MB",
        retention="7 days",
        compression="zip",
    )


# Export the configured logger
logger = _logger

__all__ = ["logger", "set_log_level", "add_file_logger", "LOG_LEVEL"]



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""
from typing import Optional, Union

import numpy as np
import torch
from transformers import pipeline, AutoModelForSpeechSeq2Seq, AutoProcessor
from parrots.log import logger

try:
    import librosa
    HAS_LIBROSA = True
except ImportError:
    HAS_LIBROSA = False

try:
    import soundfile as sf
    HAS_SOUNDFILE = True
except ImportError:
    HAS_SOUNDFILE = False

has_cuda = torch.cuda.is_available()


def load_audio(audio_path_or_bytes: Union[str, bytes], sampling_rate: int = 16000) -> np.ndarray:
    """
    Load audio file without ffmpeg dependency.
    
    Args:
        audio_path_or_bytes: Path to audio file or audio bytes
        sampling_rate: Target sampling rate
        
    Returns:
        Audio array as numpy.ndarray
    """
    if isinstance(audio_path_or_bytes, bytes):
        # Handle bytes input
        import io
        if HAS_SOUNDFILE:
            audio, sr = sf.read(io.BytesIO(audio_path_or_bytes))
            if sr != sampling_rate:
                if HAS_LIBROSA:
                    audio = librosa.resample(audio, orig_sr=sr, target_sr=sampling_rate)
                else:
                    logger.warning(f"Audio sampling rate is {sr}, but target is {sampling_rate}. "
                                   "Install librosa for automatic resampling.")
        elif HAS_LIBROSA:
            audio, sr = librosa.load(io.BytesIO(audio_path_or_bytes), sr=sampling_rate)
        else:
            raise ImportError("Either soundfile or librosa is required to load audio from bytes. "
                              "Install with: pip install soundfile librosa")
    else:
        # Handle file path input
        if HAS_LIBROSA:
            # librosa automatically resamples to target sampling_rate
            audio, sr = librosa.load(audio_path_or_bytes, sr=sampling_rate)
        elif HAS_SOUNDFILE:
            audio, sr = sf.read(audio_path_or_bytes)
            if sr != sampling_rate:
                logger.warning(f"Audio sampling rate is {sr}, but target is {sampling_rate}. "
                               "Install librosa for automatic resampling: pip install librosa")
        else:
            raise ImportError("Either librosa or soundfile is required to load audio files. "
                              "Install with: pip install soundfile librosa")
    
    # Ensure mono audio
    if len(audio.shape) > 1:
        audio = audio.mean(axis=1)
    
    # Ensure float32 dtype
    if audio.dtype != np.float32:
        audio = audio.astype(np.float32)
    
    return audio


class SpeechRecognition:
    def __init__(
            self,
            model_name_or_path: str = "BELLE-2/Belle-distilwhisper-large-v2-zh",
            use_cuda: Optional[bool] = has_cuda,
            cuda_device: Optional[int] = -1,
            max_new_tokens: Optional[int] = 128,
            chunk_length_s: Optional[int] = 15,
            batch_size: Optional[int] = 16,
            torch_dtype: Optional[str] = "float16",
            language: Optional[str] = "zh",
            ignore_warning: Optional[bool] = True,
            **kwargs
    ):
        """
        Initialize the speech recognition object.
        :param model_name_or_path: Model name or path, like:
            'BELLE-2/Belle-distilwhisper-large-v2-zh', 'distil-whisper/distil-large-v2', ...
            model in HuggingFace Model Hub and release from
        :param use_cuda: Whether or not to use CUDA for inference.
        :param cuda_device: Which cuda device to use for inference.
        :param max_new_tokens: The maximum number of new tokens to generate, ignoring the number of tokens in the
            prompt.
        :param chunk_length_s: The length in seconds of the audio chunks to feed to the model.
        :param batch_size: The batch size to use for inference.
        :param torch_dtype: The torch dtype to use for inference.
        :param language: The language of the model to use.
        :param ignore_warning: Whether to ignore the experimental warning about using chunk_length_s with seq2seq models.
            Set to True to suppress the warning. Default is True.
        :param kwargs: Additional keyword arguments passed along to the pipeline.
        """
        self.device_map = "auto"
        if use_cuda:
            if torch.cuda.is_available():
                if cuda_device == -1:
                    self.device = torch.device("cuda")
                else:
                    self.device = torch.device(f"cuda:{cuda_device}")
                    self.device_map = {"": int(cuda_device)}
            else:
                raise ValueError(
                    "'use_cuda' set to True when cuda is unavailable."
                    "Make sure CUDA is available or set `use_cuda=False`."
                )
        else:
            if torch.backends.mps.is_available():
                self.device = torch.device("mps")
                self.device_map = {"": "mps"}
            else:
                self.device = "cpu"
                self.device_map = {"": "cpu"}

        torch_dtype = (
            torch_dtype
            if torch_dtype in ["auto", None]
            else getattr(torch, torch_dtype)
        )
        
        # Prepare model loading kwargs
        model_kwargs = {
            "torch_dtype": torch_dtype,
        }
        
        self.model = AutoModelForSpeechSeq2Seq.from_pretrained(
            model_name_or_path,
            **model_kwargs
        )
        
        self.model.to(self.device)

        self.processor = AutoProcessor.from_pretrained(model_name_or_path)
        self.pipe = pipeline(
            "automatic-speech-recognition",
            model=self.model,
            tokenizer=self.processor.tokenizer,
            feature_extractor=self.processor.feature_extractor,
            device=self.device,
            torch_dtype=torch_dtype,
            max_new_tokens=max_new_tokens,
            batch_size=batch_size,
            chunk_length_s=chunk_length_s,
            ignore_warning=ignore_warning,
            **kwargs
        )
        if language == 'zh':
            self.pipe.model.config.forced_decoder_ids = (
                self.pipe.tokenizer.get_decoder_prompt_ids(
                    language=language,
                    task="transcribe"
                )
            )
        logger.info(f"Speech recognition model: {model_name_or_path} has been loaded.")

    def predict(self, inputs: Union[np.ndarray, bytes, str], preprocess_audio: bool = True):
        """语音识别用的函数，识别一个wav序列的语音
        Transcribe the audio sequence(s) given as inputs to text. See the [`AutomaticSpeechRecognitionPipeline`]
        documentation for more information.

        Args:
            inputs (`np.ndarray` or `bytes` or `str` or `dict`):
                The inputs is either :
                    - `str` that is either the filename of a local audio file. The file will be read at the correct 
                      sampling rate to get the waveform using *librosa* or *soundfile* (no ffmpeg required).
                    - `bytes` it is supposed to be the content of an audio file and is interpreted by *soundfile* 
                      or *librosa* in the same way.
                    - (`np.ndarray` of shape (n, ) of type `np.float32` or `np.float64`)
                        Raw audio at the correct sampling rate (no further check will be done)
                    - `dict` form can be used to pass raw audio sampled at arbitrary `sampling_rate` and let this
                      pipeline do the resampling. The dict must be in the format `{"sampling_rate": int, "raw":
                      np.array}` with optionally a `"stride": (left: int, right: int)` than can ask the pipeline to
                      treat the first `left` samples and last `right` samples to be ignored in decoding (but used at
                      inference to provide more context to the model). Only use `stride` with CTC models.
            preprocess_audio (`bool`, *optional*, defaults to `True`):
                Whether to preprocess audio files/bytes using librosa/soundfile instead of relying on the pipeline's
                internal ffmpeg-based processing. Set to False to use the pipeline's default behavior.
            return_timestamps (*optional*, `str` or `bool`):
                Only available for pure CTC models (Wav2Vec2, HuBERT, etc) and the Whisper model. Not available for
                other sequence-to-sequence models.

                For CTC models, timestamps can take one of two formats:
                    - `"char"`: the pipeline will return timestamps along the text for every character in the text. For
                        instance, if you get `[{"text": "h", "timestamp": (0.5, 0.6)}, {"text": "i", "timestamp": (0.7,
                        0.9)}]`, then it means the model predicts that the letter "h" was spoken after `0.5` and before
                        `0.6` seconds.
                    - `"word"`: the pipeline will return timestamps along the text for every word in the text. For
                        instance, if you get `[{"text": "hi ", "timestamp": (0.5, 0.9)}, {"text": "there", "timestamp":
                        (1.0, 1.5)}]`, then it means the model predicts that the word "hi" was spoken after `0.5` and
                        before `0.9` seconds.

                For the Whisper model, timestamps can take one of two formats:
                    - `"word"`: same as above for word-level CTC timestamps. Word-level timestamps are predicted
                        through the *dynamic-time warping (DTW)* algorithm, an approximation to word-level timestamps
                        by inspecting the cross-attention weights.
                    - `True`: the pipeline will return timestamps along the text for *segments* of words in the text.
                        For instance, if you get `[{"text": " Hi there!", "timestamp": (0.5, 1.5)}]`, then it means the
                        model predicts that the segment "Hi there!" was spoken after `0.5` and before `1.5` seconds.
                        Note that a segment of text refers to a sequence of one or more words, rather than individual
                        words as with word-level timestamps.
            generate_kwargs (`dict`, *optional*):
                The dictionary of ad-hoc parametrization of `generate_config` to be used for the generation call. For a
                complete overview of generate, check the [following
                guide](https://huggingface.co/docs/transformers/en/main_classes/text_generation).
            max_new_tokens (`int`, *optional*):
                The maximum numbers of tokens to generate, ignoring the number of tokens in the prompt.

        Return:
            `Dict`: A dictionary with the following keys:
                - **text** (`str`): The recognized text.
                - **chunks** (*optional(, `List[Dict]`)
                    When using `return_timestamps`, the `chunks` will become a list containing all the various text
                    chunks identified by the model, *e.g.* `[{"text": "hi ", "timestamp": (0.5, 0.9)}, {"text":
                    "there", "timestamp": (1.0, 1.5)}]`. The original full text can roughly be recovered by doing
                    `"".join(chunk["text"] for chunk in output["chunks"])`.

        example:
        ```
            for sample in tqdm(dataset):
                audio = sample["audio"]
                inputs = processor(audio["array"], sampling_rate=audio["sampling_rate"], return_tensors="pt")
                inputs = inputs.to(device=device, dtype=torch.float16)

                outputs = model.generate(**inputs)
                print(processor.batch_decode(outputs, skip_special_tokens=True))
        ```
        """
        # Preprocess audio input to avoid ffmpeg dependency
        if preprocess_audio and (isinstance(inputs, (str, bytes))):
            # Get the sampling rate from the feature extractor
            sampling_rate = self.processor.feature_extractor.sampling_rate
            # Load audio using librosa/soundfile instead of ffmpeg
            inputs = load_audio(inputs, sampling_rate=sampling_rate)

        return self.pipe(inputs)

    def recognize_speech_from_file(self, filename):
        """
        语音识别用的接口函数
        :param filename: 识别指定文件名的语音
        :return:
        """
        return self.predict(filename)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Cli
"""
import sys
from typing import Union

import fire
import numpy as np

sys.path.append('..')

from parrots.asr import SpeechRecognition
from parrots.tts import TextToSpeech
from parrots.log import logger

global speech_recognition_model
speech_recognition_model = None


def asr_func(audio_file_path: Union[np.ndarray, bytes, str], output_path: str = None, **kwargs):
    """
    Compute ASR result from audio file.
    """
    global speech_recognition_model
    if speech_recognition_model is None:
        speech_recognition_model = SpeechRecognition(**kwargs)
    r = speech_recognition_model.predict(audio_file_path)
    res_text = r.get("text", "")
    print(res_text, end='')  # Stdout
    if output_path:
        with open(output_path, "w") as f:
            f.write(res_text)
        logger.debug(f"ASR done, result saved: {output_path}")


global text_to_speech_model
text_to_speech_model = None


def tts_func(text: str, output_path: str, text_language: str = 'auto', **kwargs):
    """
    Compute TTS result from text.
    """
    global text_to_speech_model
    if text_to_speech_model is None:
        text_to_speech_model = TextToSpeech(**kwargs)
    text_to_speech_model.predict(
        text=text,
        text_language=text_language,
        output_path=output_path
    )
    logger.debug(f"TTS done, result saved: {output_path}")


def main():
    """Main entry point"""

    fire.Fire(
        {
            "asr": asr_func,
            "tts": tts_func,
        }
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
__version__ = '1.1.2'

from parrots.asr import SpeechRecognition
from parrots.pinyin2hanzi import Pinyin2Hanzi
from parrots.tts import TextToSpeech
from parrots.log import logger, set_log_level, add_file_logger, LOG_LEVEL


# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
from parrots import chinese_utils, japanese_utils, english_utils
from parrots.symbols import symbols

symbol_to_id = {s: i for i, s in enumerate(symbols)}


def cleaned_text_to_sequence(cleaned_text):
    """
    Converts a string of text to a sequence of IDs corresponding to the symbols in the text.
      Args:
        text: string to convert to a sequence
      Returns:
        List of integers corresponding to the symbols in the text
    """
    phones = [symbol_to_id[symbol] for symbol in cleaned_text]
    return phones


language_module_map = {"zh": chinese_utils, "ja": japanese_utils, "en": english_utils}
special = [
    # ("%", "zh", "SP"),
    ("￥", "zh", "SP2"),
    ("^", "zh", "SP3"),
]


def clean_text(text, language='zh'):
    """Clean text to sequence."""
    if language not in language_module_map:
        language = "en"
        text = " "
    for special_s, special_l, target_symbol in special:
        if special_s in text and language == special_l:
            return clean_special(text, language, special_s, target_symbol)
    language_module = language_module_map[language]
    norm_text = language_module.text_normalize(text)
    if language == "zh":
        phones, word2ph = language_module.g2p(norm_text)
        assert len(phones) == sum(word2ph)
        assert len(norm_text) == len(word2ph)
    else:
        phones = language_module.g2p(norm_text)
        word2ph = None
    for ph in phones:
        assert ph in symbols
    return phones, word2ph, norm_text


def clean_special(text, language, special_s, target_symbol):
    """
    特殊静音段sp符号处理
    """
    text = text.replace(special_s, ",")
    language_module = language_module_map[language]
    norm_text = language_module.text_normalize(text)
    phones = language_module.g2p(norm_text)
    new_ph = []
    for ph in phones[0]:
        assert ph in symbols
        if ph == ",":
            new_ph.append(target_symbol)
        else:
            new_ph.append(ph)
    return new_ph, phones[1], norm_text


if __name__ == "__main__":
    print(len(symbols))
    print(clean_text("你好%啊啊啊额、还是到付红四方。", "zh"))



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""

import os
import sys
import unittest

sys.path.append('..')
from parrots import SpeechRecognition


class TestSpeechRecognition(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.pwd_path = os.path.abspath(os.path.dirname(__file__))

    def test_recognize_speech_from_file(self):
        # 创建 SpeechRecognition 的实例
        m = SpeechRecognition()
        file_path = os.path.join(self.pwd_path, '..', 'examples', 'tushuguan.wav')
        result = m.recognize_speech_from_file(file_path)
        res = result.get("text")
        self.assertEqual(res, '北京图书馆')


if __name__ == '__main__':
    unittest.main()
